#%%
import os
import numpy as np
import xarray as xr
from datetime import datetime as dat
import pandas as pd
from dask.diagnostics import ProgressBar
import pywapor.post_et_look as post_et_look
import pywapor.general.processing_functions as PF
import pywapor.general.pre_defaults as defaults
from pywapor.general.logger import log
from pywapor.enhancers.apply_enhancers import apply_enhancer

def preprocess_func(ds):

    date_string = ds.encoding["source"].split("_")[-1]
    freq_string = ds.encoding["source"].split("_")[-2]
    unit_string = ds.encoding["source"].split("_")[-3]
    source_string = ds.encoding["source"].split("_")[-4]

    if len(date_string.split(".")) > 4:
        date = dat.strptime(date_string, '%Y.%m.%d.%H.%M.tif')
    else:
        date = dat.strptime(date_string, '%Y.%m.%d.tif')
        t_offsets = {"daily": 12, "daily-min":12, "daily-max":12} # TODO get rid of daily-min/max
        if freq_string in t_offsets.keys():
            delta = t_offsets[freq_string]
        else:
            delta = float(freq_string.split("-")[0]) * 12
        date = date + pd.Timedelta(hours = delta)

    ds = ds.drop_vars("spatial_ref")
    ds = ds.squeeze("band")
    ds = ds.drop_vars("band")
    ds = ds.expand_dims("time", axis = 0)
    ds = ds.assign_coords({"time": [date]})
    ds = ds.rename_vars({"x": f"lon", "y": f"lat"})
    ds = ds.swap_dims({"x": f"lon", "y": f"lat"})

    ds.band_data.attrs = {"unit": unit_string,
                            "source": source_string}

    return ds

def main(cmeta, dbs, epochs_info, temp_folder = None, example_ds = None,
        lean_output = True, diagnostics = {}, extra_source_enhancements = {}):
    """Create temporal composites from spatio-temporal data.

    Parameters
    ----------
    cmeta : dict
        Gives configuration option for the composite creation. Should contain at least 
        `composite_type`, `temporal_interp`, `spatial_interp` and `var_name` as keys.
    dbs : list
        list of lists, where each sublists contains the paths to geotiff files stemming
        from a single source.
    epochs_info : tuple
        Contains three lists with the index-number, starttime and endtime of each
        epoch.
    temp_folder : str, optional
        Path to folder in which temporary files are stores, providing `None` will
        force the function to do all calculations in memory, by default None.
    example_ds : xr.Dataset
        Dataset used to interpolate data (spatially), by default None.
    lean_output : bool, optional
        Whether or not to remove intermediate variables from final output, 
        used for debugging, by default True.
    diagnostics : dict, optional
        Dictionary with (`lat`, `lon`) tuples as values and (short) point descriptions/names
        as values. Several graphs are created for data at these points-of-interest, by default {}.
    extra_source_enhancements : dict, optional
        Dictionary with extra functions to apply to the data before the composites are created, 
        by default {}

    Returns
    -------
    xr.Dataset
        Dataset with the calculated composites.
    """

    source_enhancements = defaults.source_enhancements_defaults()
    source_enhancements = {**source_enhancements, **extra_source_enhancements}

    ## STEP 0: PREPARATIONS.
    if not isinstance(temp_folder, type(None)):
        fh = dbs[0][0]
        path = os.path.normpath(fh)
        if not os.path.exists(temp_folder):
            os.makedirs(temp_folder)

    if diagnostics:
        diags = {k:v for k,v in diagnostics.items()}
        if "folder" in diags.keys():
            graph_folder = diags.pop("folder")
        else:
            graph_folder = None
        labels = [None, None, None]
        log.info("--> Calculating diagnostics.")
    else:
        diags = None
        labels = ["--> Resampling datasets.",
                  "--> Interpolating {x:,} of {y:,} missing pixels.",
                  "--> Calculating composites."]

    composite_type = cmeta["composite_type"]
    temporal_interp = cmeta["temporal_interp"]
    spatial_interp = cmeta["spatial_interp"]

    styling = dict()
    markers = ["*", "o", "v", "s", "*", "p", "h"]
    colors =  ["r", "g", "b", "y", "purple", "darkblue", "gray", "orange"]
    for i, fps in enumerate(dbs):
        source = os.path.split(fps[0])[-1].split("_")[1]
        styling[i+1] = (markers[i], colors[i], 1.0, source)
    styling[999] = ("P", "orange", 1.0, "-")
    styling[0] = (".", "k", 0.7, "Interp.")
    sources = {v[3]: k for k, v in styling.items()}

    # # Check if data is static
    if np.all([isinstance(composite_type, type(False)), 
                isinstance(temporal_interp, type(False)),
                len(dbs) == 1,
                len(dbs[0]) == 1]):
        ds = xr.open_dataset(dbs[0][0], engine="rasterio")
        ds = ds.rename_vars({"x": f"lon", "y": f"lat"})
        ds = ds.swap_dims({"x": f"lon", "y": f"lat"})
        ds = ds.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)
        ds, fh_intermediate = calculate_ds(ds, None, labels[0], 
                                        encoding = {"sources": {"dtype": "uint8"}})
        # ds = ds.rename({"band": "epoch"}).assign_coords({"epoch": [-9999]})
        ds = ds.isel(band = 0)
        ds = ds.drop_vars(["band"])
        ds = ds.rename({"band_data": f"{cmeta['var_name']}"})

        unit_string = dbs[0][0].split("_")[-3]
        source_string = dbs[0][0].split("_")[-4]

        attributes = {
                        # "unit": unit_string,
                        "sources": [source_string],
                    }

        ds[cmeta['var_name']].attrs = attributes
        return ds

    ## STEP 1: COMBINE DATA AND APPLY ENHANCERS.
    ds = None
    dss = list()
    dbs_names = list()

    # Open tif-files and apply spatial interpolation
    for db in dbs:
        check_geots(db)
        sub_ds = xr.open_mfdataset(sorted(db), concat_dim = "time", engine="rasterio", combine = "nested",
                                        preprocess = preprocess_func)
        sub_ds = sub_ds.rename({"band_data": cmeta["var_name"]})

        variable = cmeta["var_name"]
        source = sub_ds[variable].source

        dbs_names.append(source)
        sub_ds["sources"] = (np.isfinite(sub_ds[variable]) * sources[source])

        if (source, variable) in source_enhancements.keys():
            for enhancer in source_enhancements[(source, variable)]:
                sub_ds, label = apply_enhancer(sub_ds, variable, enhancer, source = source, 
                                log_it = isinstance(diags, type(None)))
                sub_ds, _ = calculate_ds(sub_ds, label = label)

        if isinstance(example_ds, type(None)):
            example_ds = sub_ds.isel(time = 0).drop_vars(["time"])
        else:
            sub_ds = sub_ds.interp_like(example_ds, method = spatial_interp, kwargs={"fill_value": "extrapolate"},)

        if isinstance(diags, dict):
            # Select the pixels that contain the diagnostics points.
            sub_ds = sub_ds.sel(lat = [v[0] for v in diags.values()],
                                lon = [v[1] for v in diags.values()], method = "nearest")
            # Remove pixels that where selected twice, in case diagnostics points
            # lie within the same pixel (usually only necessary for very coarse data).
            sub_ds = sub_ds.isel(lat = np.unique(sub_ds.lat, return_index = True)[1], 
                                 lon = np.unique(sub_ds.lon, return_index = True)[1])

        dss.append(sub_ds)

    results = check_units(get_units(dss), strictness = "med") # TODO: Move to "high"
    assert np.all(results.values()), f"Combining data with different units: {get_units(dss)}"

    # This is where one dataset gets priority over another if there are sources
    # at the exact same time. That's also why we can't simply use xr.merge() here.
    for i, sub_ds in enumerate(dss):
        if i == 0:
            ds = sub_ds
        else:
            ds = ds.combine_first(sub_ds)

    all_variables = list(ds.keys())
    all_variables.remove('sources')

    for var in all_variables:
        ds[var].attrs = {"sources": dbs_names}

    if not isinstance(temp_folder, type(None)):
        fh = os.path.join(temp_folder, f"{cmeta['var_name']}_ts.nc")
    else:
        fh = None

    if temporal_interp:
        chunks = {"lat": "auto", "lon": "auto"}
    else:
        chunks = "auto"

    ds, _ = calculate_ds(ds, fh, labels[0], chunks = chunks, 
                            encoding = {"sources": {"dtype": "uint8"}})

    ## STEP 2: TEMPORAL INTERPOLATION.
    if isinstance(epochs_info, type(None)):
        return ds
    else:
        epochs, epoch_starts, epoch_ends = epochs_info

    da = xr.DataArray(epochs, coords={"time":epoch_starts})
    ds["epochs"] = da.reindex_like(ds["time"], method = "ffill", tolerance = epoch_ends[-1] - epoch_starts[-1])

    if temporal_interp:
        t_new = req_t(ds["epochs"], epochs_info, composite_type)
        ds, x, y = temporal_interpolation(ds, t_new, temporal_interp)
        if isinstance(labels[1], str):
            labels[1] = labels[1].format(x = x[all_variables[0]].values, y = y[all_variables[0]].values)
        if x[all_variables[0]].values > 0:
            ds, _ = calculate_ds(ds, fh, label = labels[1], chunks = "auto")
        else: 
            ds = ds.chunk("auto")

    da = xr.DataArray(epochs, coords={"time":epoch_starts})
    ds["epochs"] = da.reindex_like(ds["time"], method = "ffill", tolerance = epoch_ends[-1] - epoch_starts[-1])

    ## STEP 3: COMPOSITE
    for var in all_variables:
        if composite_type == "max":
            ds[f"{var}_composite"] = ds[var].groupby(ds["epochs"]).max(skipna = True).rename({"epochs": "epoch"})
        elif composite_type == "mean":
            ds[f"{var}_composite"] = ds[var].groupby(ds["epochs"]).mean(skipna = True).rename({"epochs": "epoch"})
        elif composite_type == "min":
            ds[f"{var}_composite"] = ds[var].groupby(ds["epochs"]).min(skipna = True).rename({"epochs": "epoch"})
        elif isinstance(composite_type, float):
            ds[f"{var}_composite"] = ds[var].groupby(ds["epochs"]).quantile(composite_type, skipna = True).rename({"epochs": "epoch"})
        else:
            log.warning("No valid composite_type selected.")
        ds[f"{var}_composite"].attrs = ds[var].attrs

    ds = ds.reindex({"epoch": epochs_info[0]})

    ds = ds.assign_coords(epoch_starts = ("epoch", epoch_starts))
    ds = ds.assign_coords(epoch_ends = ("epoch", epoch_ends))

    checklist = {True: dbs_names + ["Interp."], False: dbs_names}[True] # TODO interp data should not be included when not used
    sources_styling = {str(k): v for k, v in styling.items() if v[3] in checklist}
    ds.attrs = {str(k): str(v) for k, v in {**cmeta, **sources_styling}.items()}

    coords_to_keep = ["lon", "lat", "epoch", "epoch_starts", "epoch_ends", "epochs", "time"]
    coords_to_drop = [x for x in list(ds.coords) if x not in coords_to_keep]
    ds = ds.drop_vars(coords_to_drop)

    if lean_output:
        # Keep only the composites.
        ds = ds.drop_vars(all_variables + ["sources", "epochs", "time"])

    if not isinstance(temp_folder, type(None)):
        fh = os.path.join(temp_folder, f"{cmeta['var_name']}_composite.nc")
    else:
        fh = None

    ds, _ = calculate_ds(ds, fh, label = labels[2], chunks = "auto")

    if isinstance(diags, dict):
        for var in all_variables:
            post_et_look.plot_composite(ds, diags, graph_folder, band_name = var)

    # if not isinstance(fh_intermediate, type(None)):
    #     os.remove(fh_intermediate)

    return ds

def temporal_interpolation(ds, t_new, temporal_interp):
    """Fill missing and required values using temporal interpolation. In case the
    `ds` contains only a single point in time, any values at `t_new` are extrapolated
    from that single point.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset with (missing) data.
    t_new : np.ndarray
        List of extra points in time to add (and interpolate) to `ds`.
    temporal_interp : {"nearest" | "linear"}
        Type of temporal interpolation to apply.

    Returns
    -------
    xr.Dataset
        Dataset with the resampled and interpolated data.
    xr.Dataset
        Dataset with data on how many points can be interpolated (per variable).
    xr.Dataset
        Dataset with data on how many pixels are missing in total (per variable).

    """
    ds_resampled = xr.merge([ds, xr.Dataset({"time": t_new})])
    variables = list(ds_resampled.variables)
    if "sources" in variables:
        ds_resampled["sources"] = ds_resampled.sources.fillna(0.0)
        variables.remove("sources")
    y = ds_resampled.isnull()
    x = (ds_resampled.ffill("time").notnull() * ds_resampled.bfill("time").notnull() * y).sum()
    if ds.time.size > 1:
        for var in variables:
            if "lon" in ds[var].coords and "lat" in ds[var].coords:
                if x[var].values > 0:
                    ds_resampled[var] = ds_resampled[var].interpolate_na(dim = "time", method = temporal_interp)
    elif ds.time.size == 1:
        ds_resampled = ds_resampled.ffill(dim="time").bfill(dim="time")
    return ds_resampled, x, y.sum()

def calculate_ds(ds, fh = None, label = None, encoding = {}, chunks = None):
    """Operations in xarray are often done lazily and this function forces
    calculation and optionally saves the (intermediate) results to disk.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to do operation on.
    fh : str, optional
        Path to `.nc` file in which to save the calculated dataset. `fh` is never overwritten,
        but instead a "_" appended to the filename if the file already exists.
        When `None`, dataset is calculated but not written to disk, by default None.
    label : str, optional
        Label to log, by default None.
    encoding : dict, optional
        Extra encoding to apply when saving to disk, "zlib" is always applied unless
        explicitly set to False with this variable, by default {}.
    chunks : {str | dict}, optional
        Value to chunk the returned data, by default None.

    Returns
    -------
    xr.Dataset
        Returned and calculated dataset.
    str
        Path to saved `.nc` file.

    """
    if isinstance(fh, str):
        while os.path.isfile(fh): # TODO make overwrite instead, was giving weird results so did this for now.
            fh = fh.replace(".nc", "_.nc")
        if not isinstance(label, type(None)):
            log.info(label)
        folder = os.path.split(fh)[0]
        if not os.path.exists(folder):
            os.makedirs(folder)
        with ProgressBar(minimum = 30):
            varis = list(ds.variables)
            default = {"zlib": True}
            full_encoding = {v: default for v in varis if v not in list(ds.coords)}
            for var in encoding.keys():
                full_encoding[var] = {**encoding[var], **full_encoding[var]}
            ds.to_netcdf(fh, encoding = full_encoding)
        new_ds = xr.open_dataset(fh, chunks = chunks)
    else:
        if not isinstance(label, type(None)):
            log.info(label)
        with ProgressBar(minimum=30):
            new_ds = ds.compute()
    return new_ds, fh

def req_t(ds_epochs, epochs_info, composite_type):
    """Determines which epochs don't have any valid data and calculates at which time
    the dataset should be interpolated to be able to create the epochs composite.

    Parameters
    ----------
    ds_epochs : xr.DataArray
        Array with the epoch numbers for which data does currently exist.
    epochs_info : tuple
        Contains three lists with the index-number, starttime and endtime of each
        epoch.
    composite_type : str
        The type of composite determines how many point will need to be interpolated, e.g.
        to have a `mean`-composite, one point in the middle of the epoch will suffice, while
        for a `max` composite two points in time (at the very start and end of the epoch) need
        to be calculated. 

    Returns
    -------
    np.ndarray
        Array of `pd.Timestamps` at which the dataset needs to be interpolated to be able
        to create all the required composites.
    """
    missing_epochs = [x for x in np.unique(epochs_info[0]) if x not in ds_epochs]
    if composite_type == "mean":
        t_new = epochs_info[1][missing_epochs] + (epochs_info[2][missing_epochs] - epochs_info[1][missing_epochs])/2
    else:
        t1 = epochs_info[1][missing_epochs]
        t2 = epochs_info[2][missing_epochs] - pd.Timedelta(seconds=1)
        t_new = np.concatenate((t1, t2))
    return t_new

def check_geots(files):
    """Check if all `files` have the same geotransform and projection.

    Parameters
    ----------
    files : list
        List of paths to geotiff files that are required to have identical 
        geotransforms and projections.
    """
    # TODO: move to PF?
    ref = PF.get_geoinfo(files[0])
    for fh in files[1:]:
        checker = PF.get_geoinfo(fh)
        assert ref == checker, f"ERROR: {files[0]} does not have same geotransform/projection as {fh}."

def get_units(dss):
    """Get an overview of the units of different variables inside xr.Datasets.

    Parameters
    ----------
    dss : list
        List with xr.Datasets. Each variable in each dataset is checked for a 
        `"unit"` attribute.

    Returns
    -------
    dict
        Dictionary with the units per variable.
    """
    units = dict()
    for sub_ds in dss:
        variables = list(sub_ds.keys())
        for var in variables:
            if hasattr(sub_ds[var], "unit"):
                unit = sub_ds[var].unit
            else:
                unit = "unknown"
            if var in units.keys():
                units[var] = np.append(units[var], unit)
            else:
                units[var] = np.array([unit])
    return units

def check_units(units, strictness = "low"):
    """Test whether each variable (as key in `units`) has identical units.

    Parameters
    ----------
    units : dict
        Dictionary with the units per variable, can be generated with `compositer.get_units`.
    strictness : {"low" | "medium" | "high}, optional
        low - Units need to be the same, but 'unknown' units are assumed to be correct.
        med - Units need to be the same and 'unknown' units are assumed to be different.
        high - All units must be known and identical.

    Examples
    --------
    >>> units = {"test1": np.array(["C", "C","unknown"]),
    ...          "test2": np.array(["C", "C", "K"]),
    ...          "test3": np.array(["C", "C", "C"]),
    ...          "test4": np.array(["unknown", "unknown", "unknown"])}
    >>> check_units(units)
    {'test1': True, 'test2': False, 'test3': True, 'test4': True}
    >>> check_units(units, strictness = "med")
    {'test1': False, 'test2': False, 'test3': True, 'test4': True}
    >>> check_units(units, strictness = "high")
    {'test1': False, 'test2': False, 'test3': True, 'test4': False}
    """
    results = dict()
    if strictness == "low":
        for k, v in units.items():
            check = np.unique(v[v!="unknown"]).size <= 1
            results[k] = check
    if strictness == "med":
        for k, v in units.items():
            check = np.unique(v).size == 1
            results[k] = check
    if strictness == "high":
        for k, v in units.items():
            check = np.unique(v).size == 1 and "unknown" not in v
            results[k] = check
    return results

#%%

if __name__ == "__main__":

    import pywapor
    import pywapor.general.pre_defaults as defaults
    import glob

    project_folder = r"/Users/hmcoerver/pywapor_notebooks"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    startdate = "2021-07-01"
    enddate = "2021-07-11"
    epochs_info = pywapor.pre_et_look.create_dates(startdate, enddate, "DEKAD")
    var = "lulc"
    cmeta = defaults.composite_defaults()[var]

    dbs = [[r"/Users/hmcoerver/pywapor_notebooks/RAW/WAPOR/LULC_WAPOR_-_365-daily_2020.01.01.tif"]]

    temp_folder = os.path.join(project_folder, "temporary")
    example_ds = xr.open_dataset(r"/Users/hmcoerver/pywapor_notebooks/example_ds.nc")
    lean_output = True 
    diagnostics = None

#%% STEP1: COMBINE

    # ds = None
    # dss = list()
    # dbs_names = list()

    # # Open tif-files and apply spatial interpolation
    # for db in dbs:
    #     check_geots(db)
    #     sub_ds = xr.open_mfdataset(db, concat_dim = "time", engine="rasterio", combine = "nested",
    #                                     preprocess = preprocess_func)
    #     sub_ds = sub_ds.rename({"band_data": cmeta["var_name"]})

    #     variable = cmeta["var_name"]
    #     source = sub_ds[variable].source

    #     if isinstance(example_ds, type(None)):
    #         example_ds = sub_ds.isel(time = 0).drop_vars(["time"])
    #     else:
    #         sub_ds = sub_ds.interp_like(example_ds, method = "linear", kwargs={"fill_value": "extrapolate"},)

    #     dss.append(sub_ds)

    # for i, sub_ds in enumerate(dss):
    #     if i == 0:
    #         ds = sub_ds
    #     else:
    #         ds = ds.combine_first(sub_ds)

    # with ProgressBar():
    #     ds.to_netcdf(r"/Users/hmcoerver/pywapor_notebooks/step1.nc")

#%%
    # STEP 2: TIME INTERPOLATION

    # import xarray as xr
    # from dask.diagnostics import ProgressBar
    # import time

    # fh = r"/Users/hmcoerver/pywapor_notebooks/step1.nc"
    # ds = xr.open_dataset(fh, chunks = {"lat": "auto", "lon": "auto"})
    
    # new_t = req_t()

    # t1 = time.time_ns() 
    # test2 = ds.interpolate_na(dim="time", method="linear")
    # with ProgressBar():
    #     test2 = test2.compute()
    # t2 = time.time_ns()
    # print(i, (t2-t1) * 10**-9)
    
#%%

    # data = np.random.rand(1,3,3)
    # data[0,...] = np.nan
    # # data[2,2,2] = np.nan
    # # data[-1,...] = np.nan

    # ds = xr.Dataset({"ndvi": (["time", "lat","lon"], data)}, coords = {"time": np.arange(data.shape[0])})

    # x = (ds.ffill("time").notnull() * ds.bfill("time").notnull() * ds.isnull()).sum()

    # ds_interpolated = ds.interpolate_na(dim="time", method="nearest", fill_value = "extrapolate")

    # ds.ffill("time").bfill("time")
# %%
    import xarray as xr
    import os
    import numpy as np
    import glob

    # in_file = r"/Users/hmcoerver/pywapor_notebooks/temporary/lst_ts.nc"

    test_files = glob.glob(r"/Users/hmcoerver/pywapor_notebooks/temporary/*.nc")
    more_test_files = glob.glob(r"/Users/hmcoerver/pywapor_notebooks/level_1/*.nc")

    for in_file in test_files + more_test_files:
        print(os.path.split(in_file)[-1])
        ds = xr.open_dataset(in_file)

        default = {
                    # "zlib": True,
                    # "complevel": 1,
                    "scale_factor": 0.001,
                    "dtype": "int32",
                    "_FillValue": -9999
                }
        
        varis = list(ds.variables)
        encoding = {v: default for v in varis if v not in list(ds.coords)}

        out_file = r"/Users/hmcoerver/pywapor_notebooks/test.nc"
        if os.path.isfile(out_file):
            os.remove(out_file)

        ds.to_netcdf(out_file, encoding = encoding)

        print(f"size is --== {os.path.getsize(out_file)/os.path.getsize(in_file)*100:.1f}% ==-- of original")

        ds_compressed = xr.open_dataset(out_file)
        
        # print("lat_dif", ds_compressed.lat.values[:3] - ds.lat.values[:3])
            
        for var in encoding.keys():

            # print(f"{var}_dif", ds_compressed[var].values[0, :3, 0] - ds[var].values[0, :3, 0])

            if "time" in list(ds[var].coords):
                error = np.sqrt(((ds[var].isel(time=0) - ds_compressed[var].isel(time=0))**2).mean())
            elif "epoch" in list(ds[var].coords):
                error = np.sqrt(((ds[var].isel(epoch=0) - ds_compressed[var].isel(epoch=0))**2).mean())
            
            if var == "sources":
                ...
            else:
                print(error.values)

        ds = ds.close()
        ds_compressed = ds_compressed.close()

        print("")

    
# %%



import os
import time
import logging
import shutil
import fnmatch
import warnings
import concurrent.futures
import tqdm
# try:
import urllib.parse as urlparse
# except ImportError:
#     import urlparse

import requests

import itsybitsy

logger = logging.getLogger(__name__)

def copyfileobj(fsrc, fdst, callback, length=0):
    """
    https://stackoverflow.com/questions/29967487/get-progress-back-from-shutil-file-copy-thread
    """
    try:
        # check for optimisation opportunity
        if "b" in fsrc.mode and "b" in fdst.mode and fsrc.readinto:
            return _copyfileobj_readinto(fsrc, fdst, callback, length)
    except AttributeError:
        # one or both file objects do not support a .mode or .readinto attribute
        pass

    if not length:
        try:
            length = shutil.COPY_BUFSIZE
        except AttributeError:
            length = 1024*1024

    fsrc_read = fsrc.read
    fdst_write = fdst.write

    copied = 0
    while True:
        buf = fsrc_read(length)
        if not buf:
            break
        fdst_write(buf)
        # copied += len(buf)
        callback(len(buf))

def _copyfileobj_readinto(fsrc, fdst, callback, length=0):
    """readinto()/memoryview() based variant of copyfileobj().
    *fsrc* must support readinto() method and both files must be
    open in binary mode.
    """
    fsrc_readinto = fsrc.readinto
    fdst_write = fdst.write

    if not length:
        try:
            file_size = os.stat(fsrc.fileno()).st_size
        except OSError:
            file_size = 1024 * 1024
        length = min(file_size, 1024 * 1024)

    copied = 0
    with memoryview(bytearray(length)) as mv:
        while True:
            n = fsrc_readinto(mv)
            if not n:
                break
            elif n < length:
                with mv[:n] as smv:
                    fdst.write(smv)
            else:
                fdst_write(mv)
            copied += n
            callback(copied)

def _download_file(url, target, session, max_retries=3, skip_existing=True):
    """Download a single file"""
    if skip_existing and os.path.isfile(target):
        logger.debug('>>> using existing file %s', target)
        return target
    target_temp = target + '.incomplete'
    response = None
    success = False
    for n in range(max_retries):
        with session.get(url, stream=True) as response:
            total_size = int(response.headers['Content-Length'])
            with open(target_temp, "wb") as target_file:
                waitbar = tqdm.tqdm(total = total_size, position = 0, unit = "Bytes", unit_scale = True)
                callback = lambda x: waitbar.update(x)
                # shutil.copyfileobj(response.raw, target_file)
                copyfileobj(response.raw, target_file, callback)
                waitbar.close()
            data_length = response.raw.tell()
            if not data_length:
                logger.debug('Waiting 2 seconds before retry')
                time.sleep(2)
                continue
            else:
                success = True
                break
    if response is not None:
        response.raise_for_status()
    if not success:
        raise RuntimeError('Downloading {} failed after {} attempts.'.format(url, n+1))
    shutil.move(target_temp, target)
    return target


def _recursive_download(base_url, download_directory=".", username=None, password=None,
                        include=None, exclude=None, skip_existing=True,
                        download_jobs=10, download_retries=3, crawler_args=None):
    """Concurrent recursive downloader using the itsybitsy crawler

    Arguments
    ---------
    base_url : str
        Starting point for crawler
    download_directory : str
        Directory to save files to (default: current directory)
    username : str
        Username required for authentication (default: no authentication)
    password : str
        Password required for authentication (default: no authentication)
    include : list of str or str
        Download only files matching at least one of those glob patterns
        (default: download all files)
    exclude : list of str or str
        Do not download files matching at least one of those glob patterns
        (default: download all files)
    skip_existing : bool
        skip existing files
    download_jobs : int
        Number of concurrent jobs used for downloading files
    download_retries : int
        Number of retries
    crawler_args : dict
        Keyword arguments to pass to itsybitsy.crawl
    """
    logger.debug("crawling %s", base_url)

    if crawler_args is None:
        crawler_args = {}

    if username and password:
        crawler_args["auth"] = (username, password)

    crawler = itsybitsy.crawl(base_url, **crawler_args)
    base_url_normalized = next(crawler)
    base_path = urlparse.urlparse(base_url_normalized).path

    if isinstance(include, str):
        include = [include]
    if isinstance(exclude, str):
        exclude = [exclude]

    with requests.Session() as session:
        session.auth = (username, password)
        futures = set()
        with concurrent.futures.ThreadPoolExecutor(max_workers=download_jobs) as executor:
            for url in crawler:
                logger.debug("> found link: %s" % url)
                url_parts = urlparse.urlparse(url)
                file_path = url_parts.path
                if not file_path.startswith(base_path):
                    warnings.warn(
                            "File {} does not match base path {} - skipping"
                            .format(file_path, base_path))
                    continue

                target_localpath = os.path.normpath(file_path[len(base_path):])
                target_fname = os.path.basename(target_localpath)
                target_fullpath = os.path.join(download_directory, target_fname)

                if (
                        include and not
                        any(fnmatch.fnmatch(target_localpath, pattern) for pattern in include)):
                    logger.debug(">> skipping due to include pattern")
                    continue
                if (
                        exclude and
                        any(fnmatch.fnmatch(target_localpath, pattern) for pattern in exclude)):
                    logger.debug(">> skipping due to exclude pattern")
                    continue

                logger.debug(">> downloading")
                args = (url, target_fullpath, session, download_retries, skip_existing)
                future = executor.submit(_download_file, *args)
                futures.add(future)

            if futures:
                for future in concurrent.futures.as_completed(futures):
                    yield future.result()


def download_data(url, username, password, download_dir='.',
                  include='*.zip', skip_existing=True, download_jobs=10):
    """Download a URL tree recursively using itsybitsy

    Parameters
    ----------
    url : str
        Starting point for crawler
    username : str
        Username for authentication
    password  : str
        Password for authentication
    download_dir : str
        Directory to save downloaded files to (default: current directory)
    include : list of str or str
        Download only files matching at least one of those glob patterns
        (default: all .zip files)
    skip_existing : bool
        skip existing files
    download_jobs : int
        number of parallel downloads (threads)

    Returns
    -------
    generator of str
        yields paths to downloaded files
    """
    crawler_args = dict(  # passed to itsybitsy
        only_go_deeper=True,
        max_depth=None,
        max_retries=10,
        timeout=100,
        strip_fragments=True,
        max_connections=100)
    return _recursive_download(
        url,
        download_directory=download_dir,
        username=username,
        password=password,
        include=include,
        skip_existing=skip_existing,
        crawler_args=crawler_args,
        download_jobs=download_jobs
    )



import pywapor.collect as c
import tqdm
import requests
import os

def collect_sources(param, sources, dl_args, extra_source_locations = None):
    """Download an ETLook variable from selected sources.

    Parameters
    ----------
    param : str
        The variable to download.
    sources : list
        The sources from which to download `param`.
    dl_args : dict
        Should have `Dir`, `latlim`, `lonlim`, `startdate` and `enddate` as
        keys. Defining the output folder, and the spatial and temporal domain
        to be downloaded.
    extra_source_locations : dict, optional
        Describes paths to folders where non-default sources can be found, by default None.

    Returns
    -------
    list
        List of lists with each list containing the geotiff files downloaded per source.

    Examples
    --------
    >>> dl_args = {
    ...            "Dir": r"/my/folder/RAW",
    ...            "latlim": [28.9, 29.7],
    ...            "lonlim": [30.2, 31.2],
    ...            "Startdate": "2021-07-01",
    ...            "Enddate": "2021-07-03",
    ...        }
    >>> raw_ndvi_files = collect_sources("ndvi", ['MYD13', 'MOD13'], dl_args)
    >>> print(raw_ndvi_files)
    [['/my/folder/RAW/MODIS/MYD13/NDVI_MYD13Q1_-_16-daily_2021.06.18.tif'], 
    ['/my/folder/RAW/MODIS/MOD13/NDVI_MOD13Q1_-_16-daily_2021.06.26.tif']]
    """
    files = list()
    for source in sources:

        if source == "PROBAV" and param == "ndvi":
            files.append(c.PROBAV.NDVI(**dl_args))
        elif source == "MOD13" and param == "ndvi":
            files.append(c.MOD13.NDVI(**dl_args))
        elif source == "MYD13" and param == "ndvi":
            files.append(c.MYD13.NDVI(**dl_args))

        elif source == "PROBAV" and param == "r0":
            files.append(c.PROBAV.ALBEDO(**dl_args))
        elif source == "MCD43" and param == "r0":
            files.append(c.MCD43.ALBEDO(**dl_args))

        elif source == "MOD11" and param == "lst":
            files.append(c.MOD11.LST(**dl_args))
        elif source == "MYD11" and param == "lst":
            files.append(c.MYD11.LST(**dl_args))

        elif source == "CHIRPS" and param == "p_24":
            files.append(c.CHIRPS.PRECIPITATION(**dl_args))
        elif source == "SRTM" and param == "z":
            files.append([c.SRTM.DEM(**dl_args)])

        elif source == "GLOBCOVER" and param == "lulc":
            files.append([c.Globcover.LULC(**dl_args)])
        elif source == "WAPOR" and param == "lulc":
            files.append(c.WAPOR.LULC(**dl_args))

        elif source == "GEOS5" and param == "t_air_24":
            files.append(c.GEOS5.daily(**dl_args, Vars = ['t2m']))
        elif source == "GEOS5" and param == "t_air_max_24":
            files.append(c.GEOS5.daily(**dl_args, Vars = ['t2m-max']))
        elif source == "GEOS5" and param == "t_air_min_24":
            files.append(c.GEOS5.daily(**dl_args, Vars = ['t2m-min']))
        elif source == "GEOS5" and param == "u2m_24":
            files.append(c.GEOS5.daily(**dl_args, Vars = ['u2m']))
        elif source == "GEOS5" and param == "v2m_24":
            files.append(c.GEOS5.daily(**dl_args, Vars = ['v2m']))
        elif source == "GEOS5" and param == "qv_24":
            files.append(c.GEOS5.daily(**dl_args, Vars = ['qv2m']))
        elif source == "GEOS5" and param == "p_air_0_24":
            files.append(c.GEOS5.daily(**dl_args, Vars = ['slp']))

        elif source == "GEOS5" and param == "t_air_i":
            files.append(c.GEOS5.three_hourly(**dl_args, Vars = ["t2m"]))
        elif source == "GEOS5" and param == "u2m_i":
            files.append(c.GEOS5.three_hourly(**dl_args, Vars = ["u2m"]))
        elif source == "GEOS5" and param == "v2m_i":
            files.append(c.GEOS5.three_hourly(**dl_args, Vars = ["v2m"]))
        elif source == "GEOS5" and param == "qv_i":
            files.append(c.GEOS5.three_hourly(**dl_args, Vars = ["qv2m"]))
        elif source == "GEOS5" and param == "wv_i":
            files.append(c.GEOS5.three_hourly(**dl_args, Vars = ["tqv"]))
        elif source == "GEOS5" and param == "p_air_i":
            files.append(c.GEOS5.three_hourly(**dl_args, Vars = ["ps"]))
        elif source == "GEOS5" and param == "p_air_0_i":
            files.append(c.GEOS5.three_hourly(**dl_args, Vars = ["slp"]))

        elif source == "MERRA2" and param == "t_air_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['t2m']))
        elif source == "MERRA2" and param == "t_air_max_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['t2m'], data_type = ["max"]))
        elif source == "MERRA2" and param == "t_air_min_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['t2m'], data_type = ["min"]))
        elif source == "MERRA2" and param == "u2m_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['u2m']))
        elif source == "MERRA2" and param == "v2m_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['v2m']))
        elif source == "MERRA2" and param == "qv_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['q2m']))
        elif source == "MERRA2" and param == "p_air_0_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['slp']))

        elif source == "MERRA2" and param == "t_air_i":
            files.append(c.MERRA2.hourly_MERRA2(**dl_args, Vars = ["t2m"]))
        elif source == "MERRA2" and param == "u2m_i":
            files.append(c.MERRA2.hourly_MERRA2(**dl_args, Vars = ["u2m"]))
        elif source == "MERRA2" and param == "v2m_i":
            files.append(c.MERRA2.hourly_MERRA2(**dl_args, Vars = ["v2m"]))
        elif source == "MERRA2" and param == "qv_i":
            files.append(c.MERRA2.hourly_MERRA2(**dl_args, Vars = ["q2m"]))
        elif source == "MERRA2" and param == "wv_i":
            files.append(c.MERRA2.hourly_MERRA2(**dl_args, Vars = ["tpw"]))
        elif source == "MERRA2" and param == "p_air_i":
            files.append(c.MERRA2.hourly_MERRA2(**dl_args, Vars = ["ps"]))
        elif source == "MERRA2" and param == "p_air_0_i":
            files.append(c.MERRA2.hourly_MERRA2(**dl_args, Vars = ["slp"]))

        elif source == "MERRA2" and param == "ra_24":
            files.append(c.MERRA2.daily_MERRA2(**dl_args, Vars = ['swgnet']))

        elif source == "STATICS" and param == "land_mask":
            files.append(c.STATICS.collect(**dl_args, vars = ['land_mask']))
        elif source == "STATICS" and param == "lw_offset":
            files.append(c.STATICS.collect(**dl_args, vars = ['lw_offset']))
        elif source == "STATICS" and param == "lw_slope":
            files.append(c.STATICS.collect(**dl_args, vars = ['lw_slope']))
        elif source == "STATICS" and param == "r0_bare":
            files.append(c.STATICS.collect(**dl_args, vars = ['r0_bare']))
        elif source == "STATICS" and param == "r0_full":
            files.append(c.STATICS.collect(**dl_args, vars = ['r0_full']))
        elif source == "STATICS" and param == "rn_offset":
            files.append(c.STATICS.collect(**dl_args, vars = ['rn_offset']))
        elif source == "STATICS" and param == "rn_slope":
            files.append(c.STATICS.collect(**dl_args, vars = ['rn_slope']))
        elif source == "STATICS" and param == "rs_min":
            files.append(c.STATICS.collect(**dl_args, vars = ['rs_min']))
        elif source == "STATICS" and param == "t_amp_year":
            files.append(c.STATICS.collect(**dl_args, vars = ['t_amp_year']))
        elif source == "STATICS" and param == "t_opt":
            files.append(c.STATICS.collect(**dl_args, vars = ['t_opt']))
        elif source == "STATICS" and param == "vpd_slope":
            files.append(c.STATICS.collect(**dl_args, vars = ['vpd_slope']))
        elif source == "STATICS" and param == "z_obst_max":
            files.append(c.STATICS.collect(**dl_args, vars = ['z_obst_max']))
        elif source == "STATICS" and param == "z_oro":
            files.append(c.STATICS.collect(**dl_args, vars = ['z_oro']))

        elif (source, param) in extra_source_locations.keys():
            sideload_files = c.sideloader.search_product_files(source, extra_source_locations[(source, param)])
            if len(sideload_files) > 0:
                files.append(sideload_files)
        else:
            raise ValueError

        # Remove sources for which no files were downloaded.
        files = [x for x in files if len(x) > 0]

    return files

def url_to_file(url, out_file):
    """Download a file from a url.

    Parameters
    ----------
    url : str
        The url to download.
    out_file : str
        Path to where the data should be saved.
    """

    file_object = requests.get(url)
    file_object.raise_for_status()

    total_size = int(file_object.headers.get('content-length', 0))

    waitbar = tqdm.tqdm(total = total_size, unit='Bytes', unit_scale=True, position = 0)

    folder = os.path.split(out_file)[0]
    if not os.path.exists(folder):
        os.makedirs(folder)

    with open(out_file, 'wb') as z:
        for data in file_object.iter_content(chunk_size=1024):
            size = z.write(data)
            waitbar.update(size)

if __name__ == "__main__":

    project_folder = r"/Users/hmcoerver/pywapor_notebooks"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    startdate = "2021-07-01"
    enddate = "2021-07-03"

    dl_args = {
            "Dir": os.path.join(project_folder, "RAW"),
            "latlim": latlim,
            "lonlim": lonlim,
            "Startdate": startdate,
            "Enddate": enddate,
            }

    # raw_r0_files = collect_sources("r0", ['MCD43'], dl_args, None)
    # raw_lst_files = collect_sources("lst", ['MOD11', 'MYD11'], dl_args, None)
    raw_ndvi_files = collect_sources("ndvi", ['MYD13', 'MOD13'], dl_args, None)

    # ds_lst = pywapor.pre_se_root.combine_lst(raw_lst_files)

    # sds, eds, prds = pywapor.pre_se_root.calc_periods(ds_lst.time.values, "3H")

    # dl_args["Startdate"] = sds
    # dl_args["Enddate"] = eds
    # dl_args["Periods"] = prds

    # MERRAmeteo_vars = ['t2m', 'u2m', 'v2m', 'q2m',  'tpw', 'ps', 'slp']
    # GEOSmeteo_vars = ['t2m', 'u2m', 'v2m', 'qv2m', 'tqv', 'ps', 'slp']

    # # all_files = dict()
    # for var in ["t_air_i", 
    #             "u2m_i", "v2m_i", "qv_i", "wv_i", "p_air_i", "p_air_0_i"
    #             ]:

    #     files = collect_sources(var, ["GEOS5", "MERRA2"], dl_args, extra_source_locations = None)

    # print(files)



import sys
from pywapor.collect.WAPOR.DataAccess import WAPOR
import os
import glob
import pywapor
from pywapor.general.logger import log

def main(Dir, latlim, lonlim, Startdate, Enddate, Parameter, Area = None, Version = "2"):
    
    """
    This function downloads WAPOR data for the specified time
    interval, and spatial extent.

    Keyword arguments:
    -- Dir: (string) to define the folder where to save the created tiff files.
    -- Startdate: (string) defines the startdate of the required dataset in the following format "yyyy-mm-dd"
    -- Enddate: (string) defines the enddate of the required dataset in the following format "yyyy-mm-dd"
    -- Latlim: (array) defines the latitude limits in the following format [Latitude_minimum, Latitude_maximum] e.g. [10, 13]
    -- Lonlim: (array) defines the longitude limits in the following format [Longitude_minimum, Longitude_maximum] e.g. [10, 13]
    -- API_KEY: (string) this is a private API KEY that can be collected here: https://io.apps.fao.org/gismgr/api/v1/swagger-ui.html#/IAM/apiKeySignIn
    -- Parameter: (string) define the parameter that is required (possible Parameters: watertools.Collect.WAPOR.DataAccess.VariablesInfo().descriptions.keys())
    -- version: (string) default is 2, but if version 1 is required use this parameter (options are "1" or "2")

    """
    auth_token = pywapor.collect.accounts.get("WAPOR")[0]

    # print('\nDownload WAPOR %s data for period %s till %s' %(Parameter, Startdate, Enddate))
    files = WAPOR(Dir, Startdate, Enddate, latlim, lonlim, Parameter, auth_token, Area = Area, Version = Version)

    # output_folder_para = os.path.join(output_folder, "WAPOR", f"{Parameter}*.tif")
    # output_files = glob.glob(output_folder_para)

    return files

def LULC(Dir, latlim, lonlim, Startdate, Enddate, Area = None, Version = "2"):
    
    """
    This function downloads WAPOR data for the specified time
    interval, and spatial extent.

    Keyword arguments:
    -- Dir: (string) to define the folder where to save the created tiff files.
    -- Startdate: (string) defines the startdate of the required dataset in the following format "yyyy-mm-dd"
    -- Enddate: (string) defines the enddate of the required dataset in the following format "yyyy-mm-dd"
    -- Latlim: (array) defines the latitude limits in the following format [Latitude_minimum, Latitude_maximum] e.g. [10, 13]
    -- Lonlim: (array) defines the longitude limits in the following format [Longitude_minimum, Longitude_maximum] e.g. [10, 13]
    -- API_KEY: (string) this is a private API KEY that can be collected here: https://io.apps.fao.org/gismgr/api/v1/swagger-ui.html#/IAM/apiKeySignIn
    -- Parameter: (string) define the parameter that is required (possible Parameters: watertools.Collect.WAPOR.DataAccess.VariablesInfo().descriptions.keys())
    -- version: (string) default is 2, but if version 1 is required use this parameter (options are "1" or "2")

    """

    log.info("--> Downloading WAPOR.")
    auth_token = pywapor.collect.accounts.get("WAPOR")[0]

    files = WAPOR(Dir, Startdate, Enddate, latlim, lonlim, 'L1_LCC_A', auth_token, Area = Area, Version = Version)

    return files

if __name__ == '__main__':
    main(sys.argv)   


# -*- coding: utf-8 -*-
from pywapor.collect.GEOS5.DataAccess import DownloadData
import pandas as pd
import tqdm
from pywapor.general.logger import log
import numpy as np

def main(Dir, latlim, lonlim, Startdate, Enddate, Vars, Periods = [1,2,3,4,5,6,7,8], Waitbar = True):
    """
    This function downloads GEOS inst data for a given variable, time
    interval, and spatial extent.

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Vars -- ['t2m', 'v2m'] 
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
	Periods -- [1,2,3,4,5,6,7,8] Period that needs to be downloaded. 1 period is 3 hour starting from noon
    Waitbar -- 1 (Default) Will print a waitbar
    """
    no_vars = len(Vars)

    total = 0
    for sd, ed, periods in zip(Startdate, Enddate, Periods):
        no = len(pd.date_range(sd, ed, freq = "D")) * len(periods)
        total += no

    super_total = total * no_vars

    if len(Vars) == 1:
        log.info(f"--> Downloading GEOS5 (3-hourly), {Vars[0]}.")
    else:
        log.info(f"--> Downloading GEOS5 (3-hourly), {Vars}.")

    if Waitbar:
        # waitbar = tqdm.tqdm(total = no_vars * no_dates, position = 0, unit = "tiles")

        waitbar = tqdm.tqdm(desc= f"Tile: 0 / {super_total}",
                            position = 0,
                            # total=total_size,
                            unit='Bytes',
                            unit_scale=True,)

    all_files = list()

    for Var in Vars:

        for sd, ed, periods in zip(Startdate, Enddate, Periods):

            for Period in periods:

                # Download data
                files = DownloadData(Dir, Var, sd, ed, latlim, lonlim, "three_hourly", Period, waitbar)
                all_files = all_files + files


    if Waitbar:
        waitbar.close()

    all_files = np.unique(all_files).tolist()

    return all_files

if __name__ == '__main__':
    
    Dir = r"/Users/hmcoerver/pywapor_notebooks/RAW"
    latlim = [-40.0, 20.0]
    lonlim = [80.0, 100.0]
    Startdate = ["2021-07-01"]
    Enddate = ["2021-07-03"]
    Periods = [[1,2,3]]

    Vars = ['u2m']

    main(Dir, latlim, lonlim, Startdate, Enddate, Vars, Periods, Waitbar = True)




# -*- coding: utf-8 -*-
"""
Code to run the ETLook model.
"""
import os
import numpy as np
import pywapor.et_look_dev as ETLook_dev
import pywapor.et_look_v2 as ETLook_v2
import pywapor.general as g
import pywapor.general.processing_functions as PF
import pywapor.general.pre_defaults as defaults
import xarray as xr
import pandas as pd
from pywapor.general.compositer import calculate_ds
import copy

def main(input_data, et_look_version = "v2", export_vars = "default", export_to_tif = False):
    """Runs the ETLook model over the provided input data.

    Parameters
    ----------
    input_data : str | xr.Dataset
        Dataset generated by `pywapor.pre_et_look`.
    et_look_version : "v2" | "dev", optional
        Which version of the ETLook model to use, by default "v2".
    export_vars : "default" | "all" | list, optional
        Specify which variables to save inside the output file. `"Default"` stores `int_mm`,
        `t_24_mm`, `e_24_mm`, `et_24_mm`, `et_ref_24_mm`, `se_root`, `biomass_prod`,
        `epoch_ends` and `epoch_starts`. `"all"` stores all calculated variables. Use a
        list to specify a custom output set, by default "default".
    export_to_tif : bool, optional
        Return the variables selected with `export_vars` as netCDF (False) 
        or geoTIFF (True), by default False.

    Returns
    -------
    xr.Dataset | dict
        Returns a dataset or a dictionary with variable names as keys and
        lists with paths to geoTIFF files as values, depending on `export_to_tif`.
    """

    # Version
    if et_look_version == "v2":
        ETLook = ETLook_v2
        print("--> Running ETLook_v2")
    elif et_look_version == "dev":
        ETLook = ETLook_dev
        print("--> Running ETLook_dev")

    # Allow skipping of et_look-functions if not all of its required inputs are
    # available.
    g.lazifier.decorate_submods(ETLook_v2, g.lazifier.etlook_decorator)
    g.lazifier.decorate_submods(ETLook_dev, g.lazifier.etlook_decorator)

    # Inputs
    if isinstance(input_data, str):
        ds = xr.open_dataset(input_data)
    else:
        ds = input_data
        input_data = ds.encoding["source"]

    # Add constants to ds
    # ds = ds.assign(defaults.constants_defaults())

    ds = g.variables.initiate_ds(ds)

    if et_look_version == "dev":
        ds["nd_min"] = 0.1
        ds["tenacity"] = 1.0

    # Constants TODO: move to pre_et_look
    doy_epoch_start = [int(pd.Timestamp(x).strftime("%j")) for x in ds["epoch_starts"].values]
    doy_epoch_end = [int(pd.Timestamp(x).strftime("%j")) for x in ds["epoch_ends"].values]
    doy = [int((x+y)/2) for x, y in zip(doy_epoch_start, doy_epoch_end)]
    ds["doy"] = xr.DataArray(doy, coords = ds["epoch_starts"].coords)

    # ds["sc"] = ETLook.solar_radiation.seasonal_correction(ds["doy"])
    ds["decl"] = ETLook.solar_radiation.declination(ds["doy"])
    ds["iesd"] = ETLook.solar_radiation.inverse_earth_sun_distance(ds["doy"])

    ######################## MODEL ETLOOK ####################################

    # **effective_leaf_area_index*********************************************
    ds["vc"] = ETLook.leaf.vegetation_cover(ds["ndvi"], nd_min = ds["nd_min"], nd_max = ds["nd_max"], vc_pow = ds["vc_pow"])
    ds["lai"] = ETLook.leaf.leaf_area_index(ds["vc"], vc_min = ds["vc_min"], vc_max = ds["vc_max"], lai_pow = ds["lai_pow"])
    ds["lai_eff"] = ETLook.leaf.effective_leaf_area_index(ds["lai"])
    
    # *******TRANSPIRATION COMPONENT******************************************
    
    # **soil fraction*********************************************************
    ds["sf_soil"] = ETLook.radiation.soil_fraction(ds["lai"])

    # **atmospheric canopy resistance***********************************************
    ds["lat_rad"] = ETLook.solar_radiation.latitude_rad(ds["lat_deg"])
    ds["ws"] = ETLook.solar_radiation.sunset_hour_angle(ds["lat_rad"], ds["decl"])

    ds["ra_24_toa_flat"] = ETLook.solar_radiation.daily_solar_radiation_toa_flat(ds["decl"], ds["iesd"], ds["lat_rad"], ds["ws"])
    ds["trans_24"] = ETLook.solar_radiation.transmissivity(ds["ra_24"], ds["ra_24_toa_flat"])

    ds["stress_rad"] = ETLook.stress.stress_radiation(ds["ra_24"])
    ds["p_air_0_24_mbar"] = ETLook.meteo.air_pressure_kpa2mbar(ds["p_air_0_24"])
    ds["p_air_24"] = ETLook.meteo.air_pressure_daily(ds["z"], ds["p_air_0_24_mbar"])
    ds["vp_24"] = ETLook.meteo.vapour_pressure_from_specific_humidity_daily(ds["qv_24"], ds["p_air_24"])

    ds["svp_24"] = ETLook.meteo.saturated_vapour_pressure(ds["t_air_24"])
    
    if et_look_version == "dev":
        ds["svp_24"] = ETLook.meteo.saturated_vapour_pressure_average(
                    ETLook.meteo.saturated_vapour_pressure_maximum(ds["t_air_max_24"]),
                    ETLook.meteo.saturated_vapour_pressure_minimum(ds["t_air_min_24"]))

    ds["vpd_24"] = ETLook.meteo.vapour_pressure_deficit_daily(ds["svp_24"], ds["vp_24"])
    ds["stress_vpd"] = ETLook.stress.stress_vpd(ds["vpd_24"], ds["vpd_slope"])
    ds["stress_temp"] = ETLook.stress.stress_temperature(ds["t_air_24"], t_opt = ds["t_opt"], t_min = ds["t_min"], t_max = ds["t_max"])

    if et_look_version == "dev":
        if isinstance(ds["land_mask"], xr.DataArray):
            ds["rs_min"] = xr.where(ds["land_mask"] == 3, 400, 100)
        else:
            ds["rs_min"] = np.where(ds["land_mask"] == 3, 400, 100)

    ds["r_canopy_0"] = ETLook.resistance.atmospheric_canopy_resistance(ds["lai_eff"], ds["stress_rad"], ds["stress_vpd"], ds["stress_temp"], rs_min = ds["rs_min"], rcan_max = ds["rcan_max"])

    # **net radiation canopy******************************************************
    ds["t_air_k_24"] = ETLook.meteo.air_temperature_kelvin_daily(ds["t_air_24"])
    ds["l_net"] = ETLook.radiation.longwave_radiation_fao(ds["t_air_k_24"], ds["vp_24"], ds["trans_24"], vp_slope = ds["vp_slope"], vp_offset = ds["vp_offset"], lw_slope = ds["lw_slope"], lw_offset = ds["lw_offset"])
    ds["int_mm"] = ETLook.evapotranspiration.interception_mm(ds["p_24"], ds["vc"], ds["lai"], int_max = ds["int_max"])
    ds["lh_24"] = ETLook.meteo.latent_heat_daily(ds["t_air_24"])
    ds["int_wm2"] = ETLook.radiation.interception_wm2(ds["int_mm"], ds["lh_24"])
    ds["rn_24"] = ETLook.radiation.net_radiation(ds["r0"], ds["ra_24"], ds["l_net"], ds["int_wm2"])
    ds["rn_24_canopy"] = ETLook.radiation.net_radiation_canopy(ds["rn_24"], ds["sf_soil"])

    # find water region using ndvi
    if et_look_version == "dev":
        if ds.pixel_size < 250:
            if isinstance(ds["land_mask"], xr.DataArray):
                ds["land_mask"] = xr.where(ds["land_mask"] == 2, 1, ds["land_mask"])
                ds["land_mask"] = xr.where(ds["ndvi"] < 0, 2, ds["land_mask"])
            else:
                ds["land_mask"] = np.where(ds["land_mask"] == 2, 1, ds["land_mask"])
                ds["land_mask"] = np.where(ds["ndvi"] < 0, 2, ds["land_mask"])

    ds["stress_moist"] = ETLook.stress.stress_moisture(ds["se_root"], tenacity = ds["tenacity"])
    ds["r_canopy"] = ETLook.resistance.canopy_resistance(ds["r_canopy_0"], ds["stress_moist"], rcan_max = ds["rcan_max"])

    # **initial canopy aerodynamic resistance***********************************************************

    ds["z_obst"] = ETLook.roughness.obstacle_height(ds["ndvi"], ds["z_obst_max"], ndvi_obs_min = ds["ndvi_obs_min"], ndvi_obs_max = ds["ndvi_obs_max"], obs_fr = ds["obs_fr"])
    ds["z0m"] = ETLook.roughness.roughness_length(ds["lai"], ds["z_oro"], ds["z_obst"], ds["z_obst_max"], ds["land_mask"])
    ds["u_24"] = ETLook.meteo.wind_speed(ds["u2m_24"], ds["v2m_24"])
    ds["ra_canopy_init"] = ETLook.neutral.initial_canopy_aerodynamic_resistance(ds["u_24"], ds["z0m"], z_obs = ds["z_obs"])

    # **windspeed blending height daily***********************************************************
    ds["u_b_24"] = ETLook.meteo.wind_speed_blending_height_daily(ds["u_24"], z_obs = ds["z_obs"], z_b = ds["z_b"])

    # **ETLook.neutral.initial_daily_transpiration***********************************************************
    ds["ad_dry_24"] = ETLook.meteo.dry_air_density_daily(ds["p_air_24"], ds["vp_24"], ds["t_air_k_24"])
    ds["ad_moist_24"] = ETLook.meteo.moist_air_density_daily(ds["vp_24"], ds["t_air_k_24"])
    ds["ad_24"] = ETLook.meteo.air_density_daily(ds["ad_dry_24"], ds["ad_moist_24"])
    ds["psy_24"] = ETLook.meteo.psychrometric_constant_daily(ds["p_air_24"], ds["lh_24"])
    ds["ssvp_24"] = ETLook.meteo.slope_saturated_vapour_pressure_daily(ds["t_air_24"])
    ds["t_24_init"] = ETLook.neutral.initial_daily_transpiration(ds["rn_24_canopy"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_canopy"], ds["ra_canopy_init"])

    # **ETLook.unstable.initial_sensible_heat_flux_canopy_daily***********************************************************
    ds["h_canopy_24_init"] = ETLook.unstable.initial_sensible_heat_flux_canopy_daily(ds["rn_24_canopy"], ds["t_24_init"])

    # **ETLook.unstable.initial_friction_velocity_daily***********************************************************
    ds["disp"] = ETLook.roughness.displacement_height(ds["lai"], ds["z_obst"], land_mask = ds["land_mask"], c1 = ds["c1"])
    ds["u_star_24_init"] = ETLook.unstable.initial_friction_velocity_daily(ds["u_b_24"], ds["z0m"], ds["disp"], z_b = ds["z_b"])

    # **ETLook.unstable.transpiration***********************************************************
    ds["t_24"] = ETLook.unstable.transpiration(ds["rn_24_canopy"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_canopy"], ds["h_canopy_24_init"], ds["t_air_k_24"], ds["u_star_24_init"], ds["z0m"], ds["disp"], ds["u_b_24"], z_obs = ds["z_obs"], z_b = ds["z_b"], iter_h = ds["iter_h"])
    ds["t_24_mm"] = ETLook.unstable.transpiration_mm(ds["t_24"], ds["lh_24"])

    if et_look_version == "dev":
        ds["r_canopy_unstressed"] = ETLook.resistance.unstressed_canopy_resistance(ds["r_canopy"], ds["stress_moist"])
        ds["tpot_24"] = ETLook.unstable.transpiration(ds["rn_24_canopy"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_canopy_unstressed"], ds["h_canopy_24_init"], ds["t_air_k_24"], ds["u_star_24_init"], ds["z0m"], ds["disp"], ds["u_b_24"], z_obs = ds["z_obs"], z_b = ds["z_b"], iter_h = ds["iter_h"])
        ds["tpot_24_mm"] = ETLook.unstable.transpiration_mm(ds["tpot_24"], ds["lh_24"])

    #*******EVAPORATION COMPONENT****************************************************************

    # **ETLook.radiation.net_radiation_soil***********************************************************
    ds["rn_24_soil"] = ETLook.radiation.net_radiation_soil(ds["rn_24"], ds["sf_soil"])

    # **ETLook.resistance.soil_resistance***********************************************************
    ds["r_soil"] = ETLook.resistance.soil_resistance(ds["se_root"], land_mask = ds["land_mask"], r_soil_pow = ds["r_soil_pow"], r_soil_min = ds["r_soil_min"])

    # **ETLook.resistance.soil_resistance***********************************************************
    ds["ra_soil_init"] = ETLook.neutral.initial_soil_aerodynamic_resistance(ds["u_24"], z_obs = ds["z_obs"])

    # **ETLook.unstable.initial_friction_velocity_soil_daily***********************************************************
    ds["u_star_24_soil_init"] = ETLook.unstable.initial_friction_velocity_soil_daily(ds["u_b_24"], ds["disp"], z_b = ds["z_b"])

    # **ETLook.unstable.initial_sensible_heat_flux_soil_daily***********************************************************
    ds["stc"] = ETLook.radiation.soil_thermal_conductivity(ds["se_root"])
    ds["vhc"] = ETLook.radiation.volumetric_heat_capacity(ds["se_root"], porosity = ds["porosity"])    
    
    if et_look_version == "dev":
        ds["stc"] = ETLook.radiation.soil_thermal_conductivity(se_top = ds["se_top"])
        ds["vhc"] = ETLook.radiation.volumetric_heat_capacity(se_top = ds["se_top"], porosity = ds["porosity"])

    ds["dd"] = ETLook.radiation.damping_depth(ds["stc"], ds["vhc"])
    ds["g0_bs"] = ETLook.radiation.bare_soil_heat_flux(ds["doy"], ds["dd"], ds["stc"], ds["t_amp_year"], ds["lat_rad"])
    ds["g0_24"] = ETLook.radiation.soil_heat_flux(ds["g0_bs"], ds["sf_soil"], ds["land_mask"], ds["rn_24_soil"], ds["trans_24"], ds["ra_24"], ds["l_net"], ds["rn_slope"], ds["rn_offset"])
    ds["e_24_init"] = ETLook.neutral.initial_daily_evaporation(ds["rn_24_soil"], ds["g0_24"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_soil"], ds["ra_soil_init"])
    ds["h_soil_24_init"] = ETLook.unstable.initial_sensible_heat_flux_soil_daily(ds["rn_24_soil"], ds["e_24_init"], ds["g0_24"])

    # **ETLook.unstable.evaporation***********************************************************
    ds["e_24"] = ETLook.unstable.evaporation(ds["rn_24_soil"], ds["g0_24"], ds["ssvp_24"], ds["ad_24"], ds["vpd_24"], ds["psy_24"], ds["r_soil"], ds["h_soil_24_init"], ds["t_air_k_24"], ds["u_star_24_soil_init"], ds["disp"], ds["u_b_24"], z_b = ds["z_b"], z_obs = ds["z_obs"], iter_h = ds["iter_h"])
    ds["e_24_mm"] = ETLook.unstable.evaporation_mm(ds["e_24"], ds["lh_24"])
    ds["et_24_mm"] = ETLook.evapotranspiration.et_actual_mm(ds["e_24_mm"], ds["t_24_mm"])
    
    if et_look_version == "dev":
        ds["e_24_mm"] = np.clip(ds["e_24_mm"], 0, np.inf)
        ds["et_24_mm"] = np.clip(ds["et_24_mm"], 0, np.inf)

    # **ETLook.unstable.evaporation***********************************************************
    ds["rn_24_grass"] = ETLook.radiation.net_radiation_grass(ds["ra_24"], ds["l_net"], r0_grass = ds["r0_grass"])
    ds["et_ref_24"] = ETLook.evapotranspiration.et_reference(ds["rn_24_grass"], ds["ad_24"], ds["psy_24"], ds["vpd_24"], ds["ssvp_24"], ds["u_24"])
    ds["et_ref_24_mm"] = ETLook.evapotranspiration.et_reference_mm(ds["et_ref_24"], ds["lh_24"])

    if et_look_version == "dev":
        ds["lue"] = ETLook_dev.biomass.lue(ds["lue_max"], ds["stress_temp"], ds["stress_moist"], ds["eps_a"])
        ds["fpar"] = ETLook_dev.leaf.fpar(ds["vc"], ds["ndvi"])
        ds["apar"] = ETLook_dev.leaf.apar(ds["ra_24"], ds["fpar"])       
        ds["biomass_prod"] = ETLook_dev.biomass.biomass(ds["apar"], ds["lue"])         
    else:
        ds["t_air_k_min"] = ETLook.meteo.air_temperature_kelvin_daily(ds["t_air_min_24"])
        ds["t_air_k_max"] = ETLook.meteo.air_temperature_kelvin_daily(ds["t_air_max_24"])

        ds["t_air_k_12"] = ETLook.meteo.mean_temperature_kelvin_daytime(ds["t_air_k_min"], ds["t_air_k_max"])

        ds["t_dep"] = ETLook.biomass.temperature_dependency(ds["t_air_k_12"], dh_ap=ds["dh_ap"], d_s=ds["d_s"], dh_dp=ds["dh_dp"])
        ds["k_m"] = ETLook.biomass.affinity_constant_co2(ds["t_air_k_12"])
        ds["k_0"] = ETLook.biomass.inhibition_constant_o2(ds["t_air_k_12"])
        ds["tau_co2_o2"] = ETLook.biomass.co2_o2_specificity_ratio(ds["t_air_k_12"])

        ds["year"] = ds.epoch_starts.dt.year

        ds["co2_act"] = ETLook.biomass.co2_level_annual(ds["year"])
        ds["a_d"] = ETLook.biomass.autotrophic_respiration(ds["t_air_k_24"], ar_slo=ds["ar_slo"], ar_int=ds["ar_int"])

        ds["apar"] = ETLook.biomass.par(ds["ra_24"])
        ds["f_par"] = ETLook.biomass.fpar(ds["ndvi"], fpar_slope=ds["fpar_slope"], fpar_offset=ds["fpar_offset"])

        ds["co2_fert"] = ETLook.biomass.co2_fertilisation(ds["tau_co2_o2"], ds["k_m"], ds["k_0"], ds["co2_act"], o2=ds["o2"], co2_ref=ds["co2_ref"])
        ds["npp_max"] = ETLook.biomass.net_primary_production_max(ds["t_dep"], ds["co2_fert"], ds["a_d"], ds["apar"], gcgdm=ds["gcgdm"])
        ds["npp"] = ETLook.biomass.net_primary_production(ds["npp_max"], ds["f_par"], ds["stress_moist"], phot_eff=ds["phot_eff"])


    ds = ds.drop_vars([x for x in ds.variables if ds[x].dtype == object])

    fp, fn = os.path.split(input_data)

    ds = g.variables.fill_attrs(ds)
    g.network.create_network(ds, os.path.join(fp, f"et_look_{et_look_version}.html"))

    if export_vars == "all":
        ...
    elif export_vars == "default":
        keep_vars = ['int_mm',
                    't_24_mm',
                    'e_24_mm',
                    'et_24_mm',
                    'et_ref_24_mm',
                    'se_root',
                    'npp',
                    'epoch_ends',
                    'epoch_starts']
        ds = PF.ds_remove_except(ds, keep_vars)
    elif isinstance(export_vars, list):
        keep_vars = copy.copy(export_vars)
        keep_vars = np.unique(keep_vars + ['epoch_ends', 'epoch_starts']).tolist()
        ds = PF.ds_remove_except(ds, keep_vars)
    else:
        raise ValueError

    ds = ds.transpose("epoch", "lat", "lon") # set dimension order the same for all vars.
    
    fn = fn.replace("_input", "_output")
    ds, fh = calculate_ds(ds, os.path.join(fp, fn), "--> Saving outputs.")

    if export_to_tif:
        files = PF.export_ds_to_tif(ds, keep_vars, None)
        ds.close()
        os.remove(fh)
        return files
    else:
        return ds

if __name__ == "__main__":

    # project_folder = r"/Volumes/Data/FAO/WaPOR_vs_pyWaPOR/pyWAPOR_v1"
    # startdate = date = "2021-07-01"

    # level = "level_1"
    et_look_version = "v2"

    input_data = r"/Users/hmcoerver/Downloads/et_look_input.nc"
    input_data = xr.open_dataset(input_data)
    # input_data = input_data.drop_vars(["ndvi"])

    input_data = input_data.assign(defaults.constants_defaults())

    ds = main(input_data,
                et_look_version=et_look_version, 
                export_vars="default")

    # input_data["year"] = input_data.epoch_starts.dt.year





# -*- coding: utf-8 -*-
"""

Description:
This module downloads DEM data from http://earlywarning.usgs.gov/hydrodata/.
Use the DEM functions to download and create DEM images in Gtiff format.

Examples:
from pyWAPOR.Collect import SRTM
SRTM.DEM(Dir='C:/TempDEM4/', latlim=[29, 32], lonlim=[-113, -109])
"""
from pywapor.collect.SRTM.DEM import main as DEM

__all__ = ['DEM']

__version__ = '0.1'



# -*- coding: utf-8 -*-
"""
Authors: Tim Hessels
Module: WAPOR/Functions
"""

from pywapor.general import variables, processing_functions, compositer, pre_defaults, lazifier, network

__all__ = ['variables', 'processing_functions', 'compositer', 'pre_defaults', 'lazifier', 'network']

__version__ = '0.1'



from setuptools import setup, find_packages

setup(
    name = 'pywapor',
    version = '2.4.2',
    url = 'https://www.fao.org/aquastat/py-wapor/',
    author = "FAO",
    author_email = "bert.coerver@fao.org",
    license = "Apache",
    packages = find_packages(include = ['pywapor', 'pywapor.*']),
    include_package_data=True,
    python_requires='>=3.7, <3.9',
    install_requires = [
        'gdal',
        'xarray>=0.20',
        'numpy',
        'aiohttp',
        'pandas',
        'requests',
        'matplotlib',
        'netcdf4',
        'pyproj',
        'scipy',
        'fiona',
        'pycurl',
        'pyshp',
        'joblib',
        'bs4',
        'paramiko',
        'rasterio',
        'bottleneck>=1.3.1',
        'geojson',
        'vito_download',
        'nest_asyncio',
        'tqdm',
        'dask',
        'rioxarray',
        'python_log_indenter',
        'cryptography',
        'pyvis',
    ],
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)


import sys
from pywapor.collect.MCD43.DataAccess import DownloadData
import os
import glob
import numpy as np
from datetime import datetime as dat
from datetime import timedelta
import pywapor
from pywapor.general.logger import log

def main(Dir, latlim, lonlim, Startdate, Enddate, Waitbar = 1, hdf_library = None, remove_hdf = 1):
    """
    This function downloads MCD43 albedo daily data for the specified time
    interval, and spatial extent.

    Keyword arguments:
    Dir -- 'C:/file/to/path/'
    Startdate -- 'yyyy-mm-dd'
    Enddate -- 'yyyy-mm-dd'
    latlim -- [ymin, ymax]
    lonlim -- [xmin, xmax]
    username -- "" string giving the username of your NASA account (https://urs.earthdata.nasa.gov/)
    password -- "" string giving the password of your NASA account    
    Waitbar -- 1 (Default) will print a waitbar
    hdf_library -- string, if all the hdf files are already stored on computer
                    define directory to the data here
    remove_hdf -- 1 (Default), if 1 remove all the downloaded hdf files in the end    
    """
    username, password = pywapor.collect.accounts.get("NASA")

    log.info(f"--> Downloading MCD43.")
    DownloadData(Dir, Startdate, Enddate, latlim, lonlim, username, password, Waitbar, hdf_library, remove_hdf)

    # output_files = glob.glob(os.path.join(Dir, 'MODIS', 'MCD43', "*.tif"))

    sdate = dat.strptime(Startdate, "%Y-%m-%d")
    edate = dat.strptime(Enddate, "%Y-%m-%d")
    all_files = glob.glob(os.path.join(Dir, 'MODIS', 'MCD43', "*.tif"))
    start_dates = np.array([dat.strptime(os.path.split(x)[-1], "Albedo_MCD43A3_-_daily_%Y.%m.%d.tif") for x in all_files])
    end_dates = np.array([x + timedelta(days = 1) for x in start_dates])
    check = np.all([start_dates <= edate, end_dates >= sdate], axis = 0)
    all_files = np.array(all_files)[check].tolist()

    return all_files

if __name__ == '__main__':
    main(sys.argv)


"""Module containing functions that generate some default settings for `pywapor.pre_et_look`
and `pywapor.pre_se_root`.
"""
from pywapor.enhancers.temperature import lapse_rate
from functools import partial
import pywapor.enhancers.dem as dem
from pywapor.enhancers.temperature import kelvin_to_celsius
import pywapor.enhancers.lulc as lulc
from functools import partial

def constants_defaults():

    c = {
        'nd_min': 0.125,
        'nd_max': 0.8,
        'vc_pow': 0.7,
        'vc_min': 0,
        'vc_max': 0.9677324224821418,
        'lai_pow': -0.45,
        'diffusion_slope': -1.33,
        'diffusion_intercept': 1.15,
        't_min': 0,
        't_max': 50,
        'vp_slope': 0.14,
        'vp_offset': 0.34,
        'int_max': 0.2,
        'tenacity': 1.5,
        'rcan_max': 1000000,
        'ndvi_obs_min': 0.25,
        'ndvi_obs_max': 0.75,
        'obs_fr': 0.25,
        'z_obs': 2,
        'z_b': 100,
        'c1': 1,
        'iter_h': 3,
        'r_soil_pow': -2.1,
        'r_soil_min': 800,
        'se_top': 0.5,
        'porosity': 0.4,
        'r0_grass': 0.23,
        'eps_a': 0.5,

        # Biomass
        'dh_ap': 52750, 
        'd_s': 704.98, 
        'dh_dp': 211000,
        'ar_slo': 0.0, 
        'ar_int': 0.5,
        'fpar_slope': 1.257, 
        'fpar_offset': -0.161,
        'o2': 20.9,
        'co2_ref': 281,
        'gcgdm': 0.4,
        'phot_eff': 2.49,
    }

    return c


def composite_enhancements_defaults():
    """Returns a dictionary with the default functions to apply to the calculated
    composites.

    Returns
    -------
    dict
        Keys are variable names, values are lists of functions.
    """
    composite_enhancements = {
        "t_air_24":     [lapse_rate],
        "t_air_min_24": [lapse_rate],
        "t_air_max_24": [lapse_rate],
        "z":            [partial(dem.to_slope, out_var = "slope"), 
                        partial(dem.to_aspect, out_var = "aspect"),
                        partial(dem.to_lat, out_var = "lat_deg"),
                        partial(dem.to_lon, out_var = "lon_deg"),]
    }
    return composite_enhancements

def source_enhancements_defaults():
    """Returns a dictionary with the default functions to apply to source
    data.

    Returns
    -------
    dict
        Keys are tuples like `("source", "variable")`, values are lists of functions.
    """

    def remove_var(ds, var):
        return ds.drop_vars([var])

    source_enhancements = {
        ("MERRA2",  "t_air_24"):        [kelvin_to_celsius],
        ("MERRA2",  "t_air_min_24"):    [kelvin_to_celsius],
        ("MERRA2",  "t_air_max_24"):    [kelvin_to_celsius],
        ("GEOS5",   "t_air_24"):        [kelvin_to_celsius],
        ("GEOS5",   "t_air_min_24"):    [kelvin_to_celsius],
        ("GEOS5",   "t_air_max_24"):    [kelvin_to_celsius],
        ("GLOBCOVER", "lulc"):          [partial(lulc.lulc_to_x, out_var = "land_mask", 
                                                    convertor = lulc.globcover_to_land_mask()),
                                        partial(lulc.lulc_to_x, out_var = "rs_min", 
                                                    convertor = lulc.globcover_to_rs_min()),
                                        partial(lulc.lulc_to_x, out_var = "lue_max", 
                                                    convertor = lulc.globcover_to_lue_max()),
                                        partial(lulc.lulc_to_x, out_var = "z_obst_max", 
                                                    convertor = lulc.globcover_to_z_obst_max()),
                                        remove_var,
                                        ],
        ("WAPOR", "lulc"):              [partial(lulc.lulc_to_x, out_var = "land_mask", 
                                                    convertor = lulc.wapor_to_land_mask()),
                                        partial(lulc.lulc_to_x, out_var = "rs_min", 
                                                    convertor = lulc.wapor_to_rs_min()),
                                        partial(lulc.lulc_to_x, out_var = "lue_max", 
                                                    convertor = lulc.wapor_to_lue_max()),
                                        partial(lulc.lulc_to_x, out_var = "z_obst_max", 
                                                    convertor = lulc.wapor_to_z_obst_max()),
                                        remove_var,
                                        ],
    }
    return source_enhancements

def composite_defaults():
    """Returns the default composite settings for each variable.

    Returns
    -------
    dict
        Dictionary in which keys are variable names and values are dictionaries
        with `cmeta` settings.
    """

    cdefaults = {

        'ndvi': {
                    "composite_type": "mean",
                    "temporal_interp": "linear",
                    "spatial_interp": "nearest",
                    "var_name": "ndvi",
                    "var_unit": "-",
                },
        'r0': {
                    "composite_type": "mean",
                    "temporal_interp": "linear",
                    "spatial_interp": "nearest",
                    "var_name": "r0",
                    "var_unit": "-",
                },
        'lulc': {
                    "composite_type": 'mean',
                    "temporal_interp": 'nearest',
                    "spatial_interp": "linear",
                    "var_name": "lulc",
                    "var_unit": "-",
                },
        'z': {
                    "composite_type": False,
                    "temporal_interp": False,
                    "spatial_interp": "linear",
                    "var_name": "z",
                    "var_unit": "m",
                },
        'p_24': {
                    "composite_type": "mean",
                    "temporal_interp": "linear",
                    "spatial_interp": "nearest",
                    "var_name": "p_24",
                    "var_unit": "mm/day",
                },
        'se_root': {
                    "composite_type": "max",
                    "temporal_interp": False,
                    "spatial_interp": "nearest",
                    "var_name": "se_root",
                    "var_unit": "-",
                },
        'ra_24': {
                    "composite_type": "mean",
                    "temporal_interp": "linear",
                    "spatial_interp": "linear",
                    "var_name": "ra_24",
                    "var_unit": "-",
                }
        }

    meteo = {
            "composite_type": "mean",
            "temporal_interp": "linear",
            "spatial_interp": "linear",
            }

    meteo_names_units = {"t_air_max_24": "°C", "t_air_min_24":"°C", "t_air_24":"°C", 
        "u2m_24":"m/s", "v2m_24":"m/s","qv_24":"kg/kg","p_air_0_24":"kPa"}

    for name, unit in meteo_names_units.items():
        cdefaults[name] = {**{"var_name": name, "var_unit": unit}, **meteo}

    static = {
            "composite_type": False,
            "temporal_interp": False,
            "spatial_interp": "linear",    
    }

    statics_names_units = {'land_mask': "-",
                        'lw_offset': "-",
                        'lw_slope': "-",
                        'r0_bare': "-",
                        'r0_full': "-",
                        'rn_offset': "-",
                        'rn_slope': "-",
                        'rs_min': "sm-1",
                        't_amp_year': "C", 
                        't_opt': "C",
                        'vpd_slope': "mbar-1", 
                        'z_obst_max': "m",
                        'z_oro': "m",
    }

    for name, unit in statics_names_units.items():
        cdefaults[name] = {**{"var_name": name, "var_unit": unit}, **static}

    return cdefaults


# -*- coding: utf-8 -*-
"""

This module downloads daily CHIRPS 2.0 data from
ftp://chg-ftpout.geog.ucsb.edu server. The CHIRPS data is available since 1981-01-01 till the present.
The datasets will be stored in the user defined outputfolder in GEOTIFF format.


Examples:
from pyWAPOR.Collect import CHIRPS
CHIRPS.daily(Dir='C:/Temp/', Startdate='1999-02-01', Enddate='1999-02-03',
             latlim=[-10, 30], lonlim=[-20, 120])
"""

from pywapor.collect.CHIRPS.PRECIPITATION import main as PRECIPITATION

__all__ = ['PRECIPITATION']

__version__ = '0.1'
